/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include "life.h"


#define	RANDOMRADIUS	25
#define	RANDOMCHANCE	50


/*
 * List of line command routines.
 */
static	void	CmdAlgorithm(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdAlias(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdCenter(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdDestroy(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdDebug(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdEdit(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdInformation(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdDumpInitData(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdQuit(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdAdvance(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdFrequency(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdObjects(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdQuadCell(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdZero(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdRead(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdLoad(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdWrite(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdTtyInput(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdGrid(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdShape(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdInsert(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdClearPath(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdRestorePath(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdShortenPath(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdCopy(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdCopyMarked(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdBackup(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdGet(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdMove(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdMoveMarked(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdHelp(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdRule(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdEndInputLevel(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdUndo(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdRedo(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdRename(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdScales(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdSet(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdSettings(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdSize(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdVariables(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdType(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdUpdate(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdWait(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdMode(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdUnMark(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdInvertMarks(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdAssign(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdList(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdLocations(const char *, NUMBER, NUMBER, BOOL,BOOL);
static	void	CmdDefault(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdTracks(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdPrefix(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdDelay(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdRandomize(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdStop(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdForce(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdShift(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdWindow(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdSubWindow(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdHide(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdClose(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdColor(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdBorder(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdOutlines(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdStatus(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdView(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdPositioning(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdMargin(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdPrivate(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdShared(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdShow(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdJumping(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdNext(const char *, NUMBER, NUMBER, BOOL, BOOL);
static	void	CmdOverlay(const char *, NUMBER, NUMBER, BOOL, BOOL);


/*
 * Flags for line commands.
 */
#define	F_NONE			0x000	/* no special condition */
#define	F_NOARG			0x001	/* must not have an argument */
#define	F_ARG1			0x002	/* must have at least one argument */
#define	F_NORUN			0x004	/* illegal if running generations */
#define	F_STATUS		0x008	/* update status if command completes */
#define	F_UPDATE		0x010	/* update view if command completes */
#define	F_REDRAW		0x020	/* redraw screen if command completes */
#define	F_ABBR			0x040	/* abbreviation works even if ambiguous */
#define	F_FILES			0x080	/* expand file names for this command */
#define	F_BUFFERS		0x100	/* expand buffer names for this command */
#define	F_HIDE			0x200	/* hide command */


/*
 * Dispatch table for line commands.  Those commands which are ambiguous
 * must be defined so as to be contiguous in the table.  A spaces delimits
 * the command itself from the help string for the command.
 */
typedef	struct
{
	const char *	name;
	void		(*func)(const char *, NUMBER, NUMBER, BOOL, BOOL);
	int		flags;
} CMDTAB;


static const CMDTAB	cmdtab[] =
{
	{"advance (marked cells) expr (gens)", CmdAdvance, F_NORUN|F_ABBR},
	{"algorithm (for object is) type", CmdAlgorithm, F_NORUN},
	{"alias (command) name (to) value", CmdAlias, F_NONE},
	{"assign (button) num (to macro) ch", CmdAssign, F_ARG1|F_NORUN},
	{"backup (current object to base) obj", CmdBackup, F_NORUN|F_BUFFERS|F_ABBR},
	{"border (width is) width", CmdBorder, F_ARG1},
	{"center (object around) location", CmdCenter, F_NORUN|F_UPDATE},
	{"clearpath (from current object)", CmdClearPath, F_NOARG|F_NORUN},
	{"close (current window)", CmdClose, F_NOARG},
	{"color (of) type (is) color", CmdColor, F_ARG1},
	{"copy (current object to) obj", CmdCopy, F_ARG1|F_NORUN|F_BUFFERS|F_ABBR},
	{"copymarked (cells to) obj", CmdCopyMarked, F_ARG1|F_BUFFERS|F_NORUN},
	{"debug", CmdDebug, F_ARG1|F_HIDE},
	{"destroy (object named) obj", CmdDestroy, F_ARG1|F_BUFFERS|F_NORUN},
	{"default param (to) expr", CmdDefault, F_NONE},
	{"delay (between updates) expr (msec)", CmdDelay, F_STATUS|F_ABBR},
	{"dump (initialization data to) file", CmdDumpInitData, F_ARG1|F_NORUN},
	{"edit (object named) obj", CmdEdit, F_NORUN|F_REDRAW|F_BUFFERS|F_ABBR},
	{"endinputlevel", CmdEndInputLevel, F_NOARG|F_STATUS},
	{"force (running) expr (gens)", CmdForce, F_NORUN},
	{"frequency (of typeout is) expr", CmdFrequency, F_STATUS|F_ABBR},
	{"get (current object from) obj", CmdGet, F_ARG1|F_NORUN|F_BUFFERS|F_ABBR},
	{"grid (parameter) name (value) value", CmdGrid, F_NONE},
	{"help", CmdHelp, F_ABBR},
	{"hide (sub-window type) type", CmdHide, F_NONE},
	{"information (about object) obj", CmdInformation, F_REDRAW|F_BUFFERS},
	{"insert (cells from) obj", CmdInsert, F_ARG1|F_NORUN|F_BUFFERS|F_ABBR},
	{"invert (marks for all cells)", CmdInvertMarks, F_NOARG|F_NORUN|F_UPDATE},
	{"jumping (of generations is) expr", CmdJumping, F_STATUS},
	{"list (readable life files)", CmdList, F_UPDATE|F_ABBR},
	{"load (object from) file ...", CmdLoad, F_UPDATE|F_FILES|F_NORUN|F_ARG1},
	{"locations (shown are) action locations items", CmdLocations, F_NONE},
	{"margin (around edges is) value", CmdMargin, F_ARG1},
	{"mode (for movement)", CmdMode, F_NORUN|F_STATUS},
	{"move (current object to) obj", CmdMove, F_ARG1|F_NORUN|F_BUFFERS|F_ABBR},
	{"movemarked (cells to) obj", CmdMoveMarked, F_ARG1|F_NORUN|F_BUFFERS},
	{"next (object)", CmdNext, F_NOARG|F_NORUN|F_REDRAW},
	{"objects (are listed)", CmdObjects, F_ABBR},
	{"outlines (are) action", CmdOutlines, F_NONE},
	{"overlay (object) obj", CmdOverlay, F_BUFFERS|F_REDRAW},
	{"positioning (method is) method", CmdPositioning, F_NONE},
	{"prefix (for file paths is) path", CmdPrefix, F_NONE},
	{"private (window settings are) list", CmdPrivate, F_ARG1},
	{"quadcell (statistics are) action", CmdQuadCell, F_NONE},
	{"quit (program)", CmdQuit, F_NOARG|F_ABBR},
	{"randomize (live chance) percent", CmdRandomize, F_NORUN|F_UPDATE},
	{"read (object from) file", CmdRead, F_ARG1|F_FILES|F_NORUN|F_UPDATE|F_ABBR},
	{"redo expr (changes)", CmdRedo, F_NORUN|F_UPDATE},
	{"rename (current object to) obj", CmdRename, F_ARG1|F_BUFFERS|F_STATUS},
	{"restorepath (to current object)", CmdRestorePath, F_NOARG|F_NORUN},
	{"rule (for life are) born,live", CmdRule, F_NORUN|F_STATUS},
	{"scales (for viewing are) values", CmdScales, F_NONE},
	{"set (variable) name (to) expr", CmdSet, F_ARG1|F_ABBR},
	{"settings (action)", CmdSettings, F_ARG1},
	{"shape (for drawing cells is) shape", CmdShape, F_NONE},
	{"shared (window settings are) list", CmdShared, F_ARG1},
	{"shift (factor for movement is) expr", CmdShift, F_ARG1},
	{"shortenpath (from current object)", CmdShortenPath, F_NORUN},
	{"show (current window settings)", CmdShow, F_NOARG},
	{"size (for new) type (is) percentage", CmdSize, F_ARG1},
	{"status (line is) action", CmdStatus, F_NONE},
	{"stop (calculations on) condition", CmdStop, F_NORUN|F_ARG1},
	{"subwindow (using settings) name", CmdSubWindow, F_NONE},
	{"ttyinput", CmdTtyInput, F_UPDATE},
	{"tracks (for dead cells) action", CmdTracks, F_STATUS},
	{"type (value of expression) expr", CmdType, F_ARG1|F_STATUS|F_ABBR},
	{"undo expr (changes)", CmdUndo, F_NORUN|F_UPDATE|F_ABBR},
	{"unmark (all cells)", CmdUnMark, F_NOARG|F_NORUN|F_UPDATE},
	{"update (view to be current)", CmdUpdate, F_NOARG},
	{"variables (are listed)", CmdVariables, F_NOARG|F_UPDATE|F_ABBR},
	{"view (location is) parameter", CmdView, F_NONE},
	{"wait (for computations)", CmdWait, F_NOARG},
	{"window (using settings) name", CmdWindow, F_NONE},
	{"write (current object to) file", CmdWrite, F_ARG1|F_ABBR|F_FILES},
	{"zero (current object)", CmdZero, F_NOARG|F_NORUN|F_UPDATE|F_ABBR},
	{0, 0, 0}				/* ends the table */
};


/*
 * The filename prefix.
 */
static	char	prefix[MAX_PREFIX + 1];


/*
 * The table of aliased commands.
 * Each string in the table consists of two null-terminated values
 * concatenated together.  The first string is the alias name.
 * The second string is the expansion of the alias.
 */
static	char *	aliasTable[MAX_ALIASES];
static	int	aliasCount;


/*
 * Local routines.
 */
static	int	FindAlias(const char *);
static	void	BuildAliasedCommand(char *, const char *, const char *);
static	int	FindCommand(const char *, const CMDTAB **);
static	void	SplitCommand(char *, char **, char **, char **, char **);
static	void	ViewChanged(const LifeObject *);
static	void	SplitArguments(const char * args, int * argc, char *** argv);
static	const char *	ExtractArgument(const char ** args);
static	NUMBER	GetNumber(const char * cp);
static	long	GetSmallNumber(const char * cp);
static	void	SetLocationsEnabled(LifeObject *, BOOL);
static	void	SetWhereLocationsShown(LifeObject *, const char *, BOOL);
static	void	SetWhereLocationsEnabled(LifeObject *, const char *, BOOL);
static	void	CheckLocationList(const char * str);
static	LocationInfo * GetLocationInfo(LifeObject * obj, int id);


/*
 * Expand a file name.
 */
void
ExpandFileName(char * buf, int len)
{
	char *	entry;
	int	i;
	int	prefixLength;
	List	names;
	char	newPath[STRING_SIZE];

	InitList(&names);

	/*
	 * Only do filename recognition back to the last space.
	 */
	entry = strrchr(buf, ' ');

	if (entry)
		buf = entry + 1;

	/*
	 * Check the path type and get the file list according to it.
	 */
	if (buf[0] == '/')
	{
		/*
		 * The buffer contains an absolute path,
		 * so only look there.
		 */
		GetCompletionFiles(&names, "/", buf);
	}
	else if (prefix[0] == '/')
	{
		/*
		 * There was an absolute prefix set, so only
		 * collect names from there.
		 */
		GetCompletionFiles(&names, prefix, buf);
	}
	else if (prefix[0] != '\0')
	{
		/*
		 * There was a relative prefix set.
		 * Append that path to all the library directory paths.
		 */
		GetCompletionFiles(&names, prefix, buf);

		for (i = 0; i < Life.userLibCount; i++)
		{
			strcpy(newPath, Life.userLibs[i]);
			strcat(newPath, "/");
			strcat(newPath, prefix);

			GetCompletionFiles(&names, newPath, buf);
		}

		strcpy(newPath, LIFE_LIB);
		strcat(newPath, "/");
		strcat(newPath, prefix);

		GetCompletionFiles(&names, newPath, buf);
	}
	else if ((buf[0] == '.') && (buf[1] == '/'))
	{
		/*
		 * This is the current directory.
		 * Only look there.
		 */
		GetCompletionFiles(&names, ".", buf);
	}
	else if ((buf[0] == '.') && (buf[1] == '.') && (buf[2] == '/'))
	{
		/*
		 * This is the parent directory.
		 * Only look there.
		 */
		GetCompletionFiles(&names, ".", buf);
	}
	else
	{
		/*
		 * There is no prefix and this is not the current or
		 * parent directory.  Look in the library paths for files.
		 */
		GetCompletionFiles(&names, ".", buf);

		for (i = 0; i < Life.userLibCount; i++)
			GetCompletionFiles(&names, Life.userLibs[i], buf);

		GetCompletionFiles(&names, LIFE_LIB, buf);
	}

	/*
	 * If there are no files then beep and let the user fix it.
	 */
	if (names.argc == 0)
	{
		Beep();

		return;
	}

	SortList(&names);

	/*
	 * Get a pointer to the last file entry in the buffer.
	 * This is the one that we will modify based on the list results.
	 */
	entry = strrchr(buf, '/');

	if (entry)
		entry++;
	else
		entry = buf;

	/*
	 * If there is exactly one file then copy the entry into the buffer
	 * possibly along with an extra space to indicate that the name
	 * is complete.
	 */
	if (names.argc == 1)
	{
		strcpy(entry, names.argv[0]);

		if (entry[0] && (entry[strlen(entry) - 1] != '/'))
			strcat(entry, " ");

		FreeList(&names);

		return;
	}

	/*
	 * There are multiple entries in the returned list.
	 * Get the length of the common prefix for all of those entries.
	 */
	prefixLength = GetListPrefixLength(&names);

	/*
	 * If there is no prefix in common, then just display the list
	 * of entries so the user can see the one he wants.
	 */
	if (prefixLength == 0)
	{
		DisplayList(&names, FALSE);

		FreeList(&names);

		return;
	}

	/*
	 * There is a prefix, so put that many characters into the
	 * buffer so that the user can see a partial match.  Do not
	 * append a space after the entry so he knows there are multiple
	 * matches.  Show the list of files anyway so it appears up to date.
	 */
	memcpy(entry, names.argv[0], prefixLength);
	entry[prefixLength] = '\0';

	DisplayList(&names, FALSE);

	FreeList(&names);
}


/*
 * Expand a partial buffer name.
 */
void
ExpandBufferName(char * buf, int len)
{
	const LifeObject * obj;
	char *	cp;
	int		prefixLength;
	List	names;

	InitList(&names);

	/*
	 * Only do buffer recognition back to the last space.
	 */
	cp = strrchr(buf, ' ');

	if (cp)
		buf = cp + 1;

	/*
	 * Look through the list of buffers finding those which
	 * begin with the name already in the buffer.  Don't show
	 * names beginning with a dot unless the buffer also starts
	 * with a dot.
	 */
	prefixLength = strlen(buf);

	for (obj = Life.objects; obj; obj = obj->nextObject)
	{
		const char * name = GetStringBufferText(&obj->name);

		if ((name[0] == '.') && (buf[0] != '.'))
			continue;

		if ((strlen(name) >= prefixLength) &&
			(memcmp(buf, name, prefixLength) == 0))
		{
			AddListItem(&names, name);
		}
	}

 	/*
	 * If there are no matches, then beep.
	 */ 
	if (names.argc == 0)
	{
		Beep();

		return;
	}

	/*
	 * If there is exactly one match, then fill out the name
	 * with an extra space to indicate that it was exact.
	 */
	if (names.argc == 1)
	{
		strcpy(buf, names.argv[0]);

		strcat(buf, " ");

		FreeList(&names);

		return;
	}

	SortList(&names);

	/*
	 * There are multiple entries in the returned list.
	 * Get the length of the common prefix for all of those entries.
	 */
	prefixLength = GetListPrefixLength(&names);

	/*
	 * If there is no prefix in common, then just display the list
	 * of entries so the user can see the one he wants.
	 */
	if (prefixLength == 0)
	{
		DisplayList(&names, FALSE);

		FreeList(&names);

		return;
	}

	/*
	 * There is a prefix, so put that many characters into the
	 * buffer so that the user can see a partial match.  Do not
	 * append a space after the name so he knows there are multiple
	 * matches.  Display the list of buffers anyway.
	 */
	memcpy(buf, names.argv[0], prefixLength);
	buf[prefixLength] = '\0';

	DisplayList(&names, FALSE);

	FreeList(&names);
}


/*
 * Read and execute a line mode command.  This kind of command echoes,
 * and is terminated by an end of line.  Numeric arguments are available.
 * Some commands may have filename expansion done on them using a tab.
 * An escape aborts the command.
 */
void
DoLineCommand(NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;
	const CMDTAB *	cmd;		/* command structure */
	char *		initBuf;	/* initial input buffer */
	char *		name;		/* command name */
	char *		endName;
	char *		args;		/* arguments for command */
	char *		endArgs;
	const char *	aliasValue;
	int		aliasIndex;
	int		flag;		/* flags for command */
	int		matches;	/* number of matches found */
	int		len;
	int		endNameChar;
	BOOL	isPartial;
	Flags	updateFlags;
	char	prompt[64];

	/*
	 * Build a prompt depending on whether the status output is
	 * shown in our own window or not.
	 */
	if (GetStatusLocation(&window->settings) == STATUS_LOCATION_SELF)
	{
		strcpy(prompt, "Command: ");
	}
	else
	{
		sprintf(prompt, "Win[%ld] Command: ", window->id);
	}

	initBuf = "";

	/*
	 * Iterate reading the command line until the line read is not a
	 * partial one (i.e, it is finally terminated by a newline.)
	 */
	for (;;)
	{
		/*
		 * Read a full or partial command line, and return a pointer
		 * to it.  Provide the data already read from the previous
		 * reading of a partial line, if there was one.
		 */
		initBuf = ReadPartialString(prompt, initBuf, strlen(initBuf));

		/*
		 * See if this line ends in a tab, and if so remember
		 * that this is a partial line and remove the tab.
		 */
		isPartial = FALSE;
		len = strlen(initBuf);

		if ((len > 0) && (initBuf[len - 1] == '\t'))
		{
			isPartial = TRUE;
			initBuf[--len] = '\0';
		}

		/*
		 * Split the line into a command name and argument string.
		 */
		SplitCommand(initBuf, &name, &endName, &args, &endArgs);

		/*
		 * Save the terminating character and temporarily replace
		 * it with a null so that we can easily look it up.
		 */
		endNameChar = *endName;
		*endName = '\0';

		/*
		 * Look up the command name as an alias.
		 */
		aliasIndex = FindAlias(name);

		/*
		 * Restore the terminating character.
		 */
		*endName = endNameChar;

		/*
		 * If the command was an alias then process it.
		 */
		if (aliasIndex >= 0)
		{
			/*
			 * Get the alias value.
			 */
			aliasValue = aliasTable[aliasIndex];
			aliasValue += (strlen(aliasValue) + 1);

			/*
			 * Build the new command from the alias value.
			 */
			BuildAliasedCommand(initBuf, aliasValue, args);

			/*
			 * If recognition was done, then loop again
			 * showing the expansion of the alias.
			 */
			if (isPartial)
				continue;

			/*
			 * Split up the new line into a command name and
			 * its argument string.
			 */
			SplitCommand(initBuf, &name, &endName, &args, &endArgs);
		}

		/*
		 * Save the terminating character and temporarily replace
		 * it with a null so that we can easily look it up.
		 */
		endNameChar = *endName;
		*endName = '\0';

		/*
		 * Find the command name and remember which one this is.
		 * We don't yet cause an error for invalid commands
		 * since we may be in a partial command which could
		 * still be fixed up.
		 */
		matches = FindCommand(name, &cmd);

		/*
		 * Restore the character terminating the command.
		 */
		*endName = endNameChar;

		/*
		 * If the line is not a partial line, then go execute it.
		 */
		if (!isPartial)
			break;

		/*
		 * This is a partial command line.
		 * If the command is valid and doesn't yet have a space
		 * after it, then complete the command name and add a
		 * following space to indicate that the command is ok.
		 */
		if (*name && (matches == 1) && (endNameChar == '\0'))
		{
			strcpy(name, cmd->name);

			endName = strchr(name, ' ');

			if (endName)
				*endName = '\0';

			strcat(name, " ");

			continue;
		}

		/*
		 * If the command is valid and the command allows filename
		 * recognition, then do that.
		 */
		if ((matches == 1) && (cmd->flags & F_FILES))
		{
			ExpandFileName(args, STRING_SIZE - (args - initBuf));

			continue;
		}

		/*
		 * If the command is valid and the command allows object
		 * recognition, then do that.
		 */
		if ((matches == 1) && (cmd->flags & F_BUFFERS))
		{
			ExpandBufferName(args, STRING_SIZE - (args - initBuf));

			continue;
		}

		/*
		 * The command is illegal or else it doesn't want recognition.
		 * Just beep and ignore the tab character.  The user can
		 * then edit the line to fix the problem.
		 */
		Beep();
	}

	/*
	 * The command line is completed.
	 * If it is empty then just ignore it but make sure that
	 * text mode is disabled.
	 */
	if (*name == '\0')
	{
		Life.device->setViewModeIsText(Life.device, FALSE);

		return;
	}

	/*
	 * Complain about an invalid command if necessary.
	 */
	if (matches == 0)
		Error("Unknown line command");

	if (matches > 1)
		Error("Ambiguous line command");

	/*
	 * Delete the trailing blanks from the argument string.
	 */
	*endArgs = '\0';

	/*
	 * Get the flags for the command and verify some things.
	 */
	flag = cmd->flags;

	if (flag & F_NORUN)
		CheckRun();

	if ((flag & F_ARG1) && (*args == '\0'))
		Error("Missing argument");

	if ((flag & F_NOARG) && *args)
		Error("Argument not allowed");

	/*
	 * Call the function with the arguments.
	 */
	cmd->func(args, arg1, arg2, got1, got2);

	/*
	 * Set update flags as specified by the command.
	 */
	updateFlags = 0;

	if (flag & F_STATUS)
		updateFlags |= U_STATUS;

	if (flag & F_UPDATE)
		updateFlags |= U_ALL;

	if (flag & F_REDRAW)
		updateFlags |= U_REDRAW;

	AddWindowObjectUpdateFlags(Life.currentObject, updateFlags);

	Life.device->setViewModeIsText(Life.device, FALSE);
}


/*
 * Search the specified string for an alias command match.
 * Returns the index of the found alias entry, or -1 if the
 * alias is not defined.
 */
static int
FindAlias(const char * name)
{
	int	index;

	/*
	 * Search the alias table entries.
	 */
	for (index = 0; index < aliasCount; index++)
	{
		if (strcmp(name, aliasTable[index]) == 0)
			return index;
	}

	/*
	 * The alias is not defined.
	 */
	return -1;
}


/*
 * Full in the initial buffer with the specified alias value
 * concatenated with the specified additional arguments (if any).
 * This routine knows that the extra arguments are contained within
 * the initial buffer, and so we copy the new line into a temporary
 * buffer before building it in the final location.
 */
static void
BuildAliasedCommand(char * initBuf, const char * alias,
	const char * extraArgs)
{
	char	tempBuf[STRING_SIZE];

	/*
	 * If there were no extra arguments then we can just copy the
	 * alias into the buffer with no changes.
	 */
	if (*extraArgs == '\0')
	{
		strcpy(initBuf, alias);

		return;
	}

	/*
	 * Make sure the expanded line will not be too long.
	 * The extra characters count the space separating the two
	 * parts and the ending null.
	 */
	if (strlen(alias) + strlen(extraArgs) + 2 > STRING_SIZE)
		Error("Aliased command too long");

	/*
	 * Concatencate the strings into our temp buffer.
	 */
	strcpy(tempBuf, alias);
	strcat(tempBuf, " ");
	strcat(tempBuf, extraArgs);

	/*
	 * Copy the result back into the initial buffer,.
	 */
	strcpy(initBuf, tempBuf);
}


/*
 * Examine the string in the indicated buffer to find the location of a
 * possible command name and possible arguments which follow the command.
 * Leading and trailing blanks are ignored for both of these values.
 * However, the argument string is allowed to contain embedded blanks
 * and therefore can represent multiple values.  Missing values return
 * values which point to an empty strings.  The returned values point into
 * the data in the supplied buffer.  The data in the supplied buffer
 * is not modified.
 */
static void
SplitCommand(char * buf, char ** retName, char ** retEndName,
	char ** retArgs, char ** retEndArgs)
{
	/*
	 * Skip leading blanks to find the beginning of the
	 * command name (if present).
	 */
	while (isBlank(*buf))
		buf++;

	/*
	 * Save the command name.
	 */
	*retName = buf;

	/*
	 * Find the end of the command name.
	 */
	while (*buf && !isBlank(*buf))
		buf++;

	/*
	 * Save the end of the command name.
	 */
	*retEndName = buf;

	/*
	 * Skip blanks to find the beginning of the arguments (if present).
	 */
	while (isBlank(*buf))
		buf++;

	/*
	 * Save the beginning of the arguments.
	 */
	*retArgs = buf;

	/*
	 * If there are no arguments then set the end of the arguments
	 * to the same place.
	 */
	if (*buf == '\0')
	{
		*retEndArgs = buf;

		return;
	}

	/*
	 * Skip to the end of the arguments.
	 */
	buf += strlen(buf);

	/*
	 * Back up to the first trailing blank, if any.
	 */
	while (isBlank(buf[-1]))
		buf--;

	/*
	 * Save the end of the arguments.
	 */
	*retEndArgs = buf;
}


/*
 * Search the specified string for a command match, and indirectly return
 * a pointer to the found command, if any.  Returns the number of commands
 * which match.  Zero indicates no match was found, and more than one
 * indicates the match was ambiguous.
 */ 
static int
FindCommand(const char * name, const CMDTAB ** retCmd)
{
	const CMDTAB *	cmd;		/* command structure */
	const CMDTAB *	cmdWorks;	/* command that was found */
	const CMDTAB *	cmdBest;	/* command found that is wanted */
	int		matches;	/* number of matches found */

	*retCmd = 0;
	matches = 0;

	/*
	 * Seatch the table for the command.
	 * Exact match always wins, followed by an indicated abbreviation,
	 * followed by a unique abbreviation.
	 */
	cmdBest = 0;
	cmdWorks = 0;

	for (cmd = cmdtab; cmd->name != 0; cmd++)
	{
		if (strcmp(name, cmd->name) == 0)
		{
			cmdBest = cmd;

			break;
		}

		if (!IsAbbreviation(name, cmd->name))
			continue;

		cmdWorks = cmd;
		matches++;

		if (cmd->flags & F_ABBR)
			cmdBest = cmd;
	}

	/*
	 * If we found an exact match or an abbreviation of an
	 * important command, then return that one unambiguously.
	 */
	if (cmdBest)
	{
		*retCmd = cmdBest;

		return 1;
	}

	/*
	 * Return the found command if any, and the number of matches.
	 */
	*retCmd = cmdWorks;

	return matches;
}


/*
 * Extract the next argument from the specified string if any,
 * and advance the string pointer past the argument to the beginning
 * of the next argument, if any.  The argument can be surrounded by
 * spaces.  Returns an empty string if there is no argument.
 * The returned string data is static and is overwritten on each call.
 */
static const char *
ExtractArgument(const char ** argString)
{
	static char argData[STRING_SIZE + 1];
	const char * beginArg;
	const char * endArg;
	int len;

	/*
	 * Look for the begining of the argument.
	 */
	beginArg = *argString;

	while (*beginArg == ' ')
		beginArg++;

	/*
	 * If there is no argument then update the
	 * argument pointer and return an empty string.
	 */
	if (*beginArg == '\0')
	{
		*argString = beginArg;

		return "";
	}

	/*
	 * Find the end of the argument.
	 */
	endArg = beginArg;

	while (*endArg && (*endArg != ' '))
		endArg++;

	/*
	 * Copy the argument into the static buffer.
	 */
	len = (endArg - beginArg);

	if (len >= STRING_SIZE)
		Error("Command line too long");

	memcpy(argData, beginArg, len);
	argData[len] = '\0';

	/*
	 * Skip trailing spaces.
	 */
	while (*endArg == ' ')
		endArg++;

	/*
	 * Update the pointer and return the address
	 * of the buffer.
	 */
	*argString = endArg;

	return argData;
}


/*
 * Split the argument string into individual arguments according
 * to the spaces within the string.  Returns the number of arguments
 * and a pointer to the table of argument string pointers.  The
 * table of strings is also terminated by a null pointer entry.
 * The string data is static and is overwritten on each call.
 */
static void
SplitArguments(const char * argString, int * argc, char *** argv)
{
	static char argData[STRING_SIZE + 1];
	static char * argTable[ARGS_SIZE + 1];
	char * cp;
	int argCount;
	int len;

	/*
	 * Copy the string into our static buffer.
	 */
	len = strlen(argString);

	if (len > STRING_SIZE)
		Error("Command line too long");

	memcpy(argData, argString, len);
	argData[len] = '\0';

	/*
	 * Parse the buffer into arguments.
	 */
	argCount = 0;
	cp = argData;

	while (TRUE)
	{
		/*
		 * Skip spaces and if a null is seen break out.
		 */
		while (*cp == ' ')
			cp++;

		if (*cp == '\0')
			break;

		/*
		 * This is the beginning of another argument.
		 * Store the pointer to it as the next entry
		 * in the argument table.
		 */
		if (argCount >= ARGS_SIZE)
			Error("Too many command line arguments");

		argTable[argCount++] = cp;

		/*
		 * Skip to the end of the argument by looking
		 * for a space or null character.  If a space
		 * is found then null terminate the argument
		 * and advance over it.
		 */
		while (*cp && (*cp != ' '))
			cp++;

		if (*cp)
			*cp++ = '\0';	
	}

	/*
	 * Terminate the argument list with a null pointer.
	 */
	argTable[argCount] = 0;

	/*
	 * Return the results.
	 */
	*argc = argCount;
	*argv = argTable;
}


/*
 * Show the defined aliases.
 */
static void
ShowAliases(void)
{
	const char *	alias;
	const char *	value;
	int		index;

	if (aliasCount == 0)
		ShowHelp("There are no aliases defined.\n");
	else
	{
		ShowHelp("Alias      Value\n");
		ShowHelp("-----      -----\n");

		for (index = 0; index < aliasCount; index++)
		{
			alias = aliasTable[index];
			value = alias + strlen(alias) + 1;

			FormatHelp("%-10s ", alias);
			ShowHelp(value);
			ShowHelp("\n");
		}
	}

	ShowHelp("\n");
	EndHelp(TRUE);
}


/*
 * Define the algorithm to be used for the object.
 * This will convert the format of the cells if possible.
 */
static void
CmdAlgorithm(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;

	/*
	 * If there are no arguments then show the current algorithm.
	 */
	if (*cp == '\0')
	{
		cp = GetAlgorithmName(obj->algorithm);

		if (cp == 0)
			Error("Unknown algorithm name");

		Message(cp);

		return;
	}

	LifeAlgorithm * algorithm = GetAlgorithmByName(cp);

	if (algorithm == 0)
		Error("Invalid algorithm name");

	if (!SetObjectAlgorithm(obj, algorithm))
		Error("Algorithm could not be changed (object too large)");
}


/*
 * Define an alias for a command.
 */
static void
CmdAlias(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	const char *	cp2;
	char *		alias;
	char *		value;
	char *		value2;
	int		index;

	/*
	 * If there are no arguments then show the list of aliases.
	 */
	if (*cp == '\0')
	{
		ShowAliases();

		return;
	}

	/*
	 * Verify that there is an alias name and an alias value.
	 */
	cp2 = cp;

	while (*cp2 && !isBlank(*cp2))
		cp2++;

	while (isBlank(*cp2))
		cp2++;

	if (*cp2 == '\0')
		Error("Missing alias value");

	/*
	 * Allocate a string for the alias and its value.
	 */
	alias = strdup(cp);

	if (alias == 0)
		Error("Cannot allocate alias");

	/*
	 * Split off the alias value and null terminate the alias name.
	 */
	value = alias;

	while (!isBlank(*value))
		value++;

	*value++ = '\0';

	value2 = value;

	while (isBlank(*value2))
		value2++;

	while (*value2)
		*value++ = *value2++;

	*value = '\0';

	/*
	 * Remove trailing blanks from the alias value.
	 */
	while (isBlank(value[-1]))
		value--;

	*value = '\0';

	/*
	 * See if the alias is already defined.
	 * If so, then just replace it.
	 */
	index = FindAlias(alias);

	if (index >= 0)
	{
		free(aliasTable[index]);
		aliasTable[index] = alias;

		return;
	}

	/*
	 * This is a new alias.
	 * Try to add it to the alias table.
	 */
	if (aliasCount >= MAX_ALIASES)
	{
		free(alias);

		Error("Too many aliases");
	}

	aliasTable[aliasCount] = alias;
	aliasCount++;
}


/*
 * Advance marked cells by the specified number of generations.
 * This uses the temporary object to run the marked object.
 */
static void
CmdAdvance(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;
	LifeObject * tempObj = Life.tempObject;
	Count count = 1;

	if (*cp)
	{
		count = GetSmallExpression(cp);

		if (count <= 0)
			return;
	}

	if (obj == tempObj)
		Error("Cannot advance temporary object");

	Backup();

	ClearSelection(obj);

	if (!SelectMarkedCells(obj))
		Error("No marked cells to advance");

	MoveSelection(obj, tempObj);

	tempObj->genLeft = count;
	tempObj->isTracksUsed = obj->isTracksUsed;

	while ((tempObj->genLeft > 0) && !Life.stop)
		DoGenerations(tempObj);

	AddObject(tempObj, obj, ADD_RELATIVE);
}


/*
 * Assign a button to a macro.
 */
static void
CmdAssign(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Value	button;
	int	macro;

	button = 0;

	while (isDigit(*cp))
		button = button * 10 + *cp++ - '0';

	if (*cp && !isBlank(*cp))
		Error("Illegal button value");

	while (isBlank(*cp))
		cp++;

	macro = *cp;

	if (macro)
		cp++;

	while (isBlank(*cp))
		cp++;

	if (*cp || ((macro != '\0') && !MacroIsValid(macro)))
		Error("Illegal macro name");

	if (!Life.device->assign(Life.device, button, macro))
		Error("Button is not assignable");
}


/*
 * Clear the path from the current object.
 */
static void
CmdClearPath(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Backup();

	ClearPath(Life.currentObject);
}


/*
 * Shorten the path from the current object by one or more segments.
 */
static void
CmdShortenPath(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Count count = 1;

	if (*cp)
		count = GetSmallNumber(cp);

	if (count <= 0)
		Error("Bad shorten count value");

	Backup();

	ShortenPath(Life.currentObject, count);
}


/*
 * Restore the path that was set for an object.
 */
static void
CmdRestorePath(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Backup();

	if (!RestorePath(Life.currentObject))
		Error("There is no path to restore");
}


/*
 * Copy the current object into another object.
 */
static void
CmdCopy(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	SaveCurrentViewParameters();

	CopyObject(Life.currentObject, GetObject(cp));
}


/*
 * Backup the current object into a new object.
 * Note: This is not the same thing as the undo capability.
 */
static void
CmdBackup(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	SaveCurrentViewParameters();
	BackupObject(Life.currentObject, cp);
}


/*
 * Copy the currently marked cells into another object.
 */
static void
CmdCopyMarked(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	CopyMarkedObject(Life.currentObject, GetObject(cp));
}


/*
 * Get the current object from another object.
 * This totally destroys the current object.
 */
static void
CmdGet(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = FindObject(cp);

	if (obj == 0)
		Error("No such object");

	CopyObject(obj, Life.currentObject);

	RestoreCurrentViewParameters();
}


/*
 * Center the current object either around the cursor location
 * or around the origin.  The marks, tracked cells, where locations,
 * and path are moved with the object.  But the origin is set to the
 * true origin or to the cursor location.
 */
static void
CmdCenter(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;

	NUMBER centerRow;
	NUMBER centerCol;

	if ((*cp == '\0') || IsAbbreviation(cp, "origin"))
	{
		centerRow = numZero();
		centerCol = numZero();
	}
	else if (IsAbbreviation(cp, "cursor"))
	{
		centerRow = obj->currentInfo.row;
		centerCol = obj->currentInfo.col;
	}
	else
		Error("Center must be origin or cursor");

	CenterObject(obj, centerRow, centerCol);

	SetLocation(obj, &obj->originInfo, centerRow, centerCol);
}


/*
 * Set a default value for new objects or windows.
 */
static void
CmdDefault(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;
	const char *	param;
	Value		val;

	param = ExtractArgument(&cp);

	val = 1;

	if (IsAbbreviation(param, "scale"))
	{
		if (*cp)
			val = GetSmallExpression(cp);

		if ((val == 0) || (val > MAX_SCALE))
			Error("Bad scale value");

		if (val == -1)
			val = 1;

		SetDefaultScale(&window->settings, val);

		return;
	}

	if (IsAbbreviation(param, "frequency"))
	{
		if (*cp)
			val = GetSmallExpression(cp);

		if (val <= 0)
			Error("Illegal frequency value");

		Life.defaultFrequency = val;
		Life.defaultIsJumpingEnabled = FALSE;

		return;
	}

	if (IsAbbreviation(param, "jumping"))
	{
		if (*cp)
			val = GetSmallExpression(cp);

		if (val <= 0)
			Error("Illegal jumping value");

		Life.defaultFrequency = val;
		Life.defaultIsJumpingEnabled = (val > 1);

		return;
	}

	if (IsAbbreviation(param, "tracks"))
	{
		if (IsAbbreviation(cp, "on"))
			Life.defaultTracks = TRUE;
		else if (IsAbbreviation(cp, "off"))
			Life.defaultTracks = FALSE;
		else
			Error("Bad tracks default");

		return;
	}

	if (IsAbbreviation(param, "algorithm"))
	{
		if (*cp == 0)
			Error("No algorithm given");

		LifeAlgorithm * algorithm = GetAlgorithmByName(cp);

		if (algorithm == 0)
			Error("Unknown algorithm");

		Life.defaultAlgorithm = algorithm;

		return;
	}

	if (IsAbbreviation(param, "sharedSettings"))
	{
		int argCount;
		char ** argTable;
		SettingsFlags flags;

		/*
	 	 * Split the arguments.
	 	 */
		SplitArguments(cp, &argCount, &argTable);

		/*
		 * Parse the settings list.
		 */
		flags = ParseSettingsFlags(argCount, argTable);

		if (flags == 0)
			Error("Invalid settings name");

		Life.defaultSharedSettings = flags;

		return;
	}

	Error("Default must be algorithm, scale, frequency, jumping, tracks, or sharedSettings");
}


/*
 * Force the running of generations despite the presence of any
 * stop conditions.
 */
static void
CmdForce(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;
	Count	count = 1;
	BOOL	stopWasEnabled;

	if (*cp)
	{
		count = GetSmallExpression(cp);

		if (count <= 0)
			return;
	}

	Backup();

	/*
	 * Save the current stop enabled flag, clear it,
	 * manually run the specified number of generations,
	 * then restore the stop enabled flag.
	 */
	stopWasEnabled = obj->stopIsEnabled;

	obj->stopIsEnabled = FALSE;
	obj->genLeft = count;

	while (!Life.stop && (obj->genLeft > 0))
		DoGenerations(obj);

	if (obj->stopCell)
	{
		obj->stopState = !IsCellSet(obj, obj->stopRow, obj->stopCol);
	}

	obj->stopIsEnabled = stopWasEnabled;

	UpdateAllWindows();
}


/*
 * Set or clear a stop condition.  This is things like a cell changing
 * state, or a cell being inside, outside, or on the current path.
 */
static void
CmdStop(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;
	const char * command;
	Value		value;

	command = ExtractArgument(&cp);

	/*
	 * Now check for the various flavors of stopping.
	 */
	if (IsAbbreviation(command, "none"))
	{
		obj->stopIsEnabled = FALSE;
		obj->stopGap = 0;
		obj->stopOnPath = FALSE;
		obj->stopInsidePath = FALSE;
		obj->stopOutsidePath = FALSE;
		obj->stopCell = FALSE;
		obj->stopState = FALSE;
		obj->stopRow = numZero();
		obj->stopCol = numZero();
		obj->stopExpr[0] = '\0';

		return;
	}

	if (IsAbbreviation(command, "enabled"))
	{
		obj->stopIsEnabled = TRUE;

		return;
	}

	if (IsAbbreviation(command, "disabled"))
	{
		obj->stopIsEnabled = FALSE;

		return;
	}

	if (IsAbbreviation(command, "cell"))
	{
		obj->stopIsEnabled = TRUE;
		obj->stopCell = TRUE;
		obj->stopRow = obj->currentInfo.row;
		obj->stopCol = obj->currentInfo.col;

		obj->stopState = !IsCellSet(obj, obj->stopRow, obj->stopCol);

		return;
	}

	if (IsAbbreviation(command, "on-path"))
	{
		obj->stopIsEnabled = TRUE;
		obj->stopOnPath = TRUE;

		return;
	}

	if (IsAbbreviation(command, "inside-path"))
	{
		obj->stopIsEnabled = TRUE;
		obj->stopInsidePath = TRUE;

		return;
	}

	if (IsAbbreviation(command, "outside-path"))
	{
		obj->stopIsEnabled = TRUE;
		obj->stopOutsidePath = TRUE;

		return;
	}

	if (IsAbbreviation(command, "gap"))
	{
		value = GetSmallExpression(cp);

		if (value <= 0)
			Error("Bad gap value");

		if (value > MAX_GAP)
			Error("Gap value too large");

		obj->stopIsEnabled = TRUE;
		obj->stopGap = value;

		return;
	}

	if (IsAbbreviation(command, "expression"))
	{
		if (strlen(cp) >= sizeof(obj->stopExpr))
			Error("Stop expression too long");

		value = GetSmallExpression(cp);

		obj->stopIsEnabled = TRUE;
		strcpy(obj->stopExpr, cp);

		return;
	}

	Error("Bad stop argument");
}


/*
 * Command to set jumping value.
 * This is similar to the frequency command, but multiple generations
 * are computed at one time instead of individually.
 */
static void
CmdJumping(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Count freq = 1;

	if (*cp)
		freq = GetSmallExpression(cp);

	if (freq <= 0)
		Error("Illegal jumping value");

	if (freq > MAX_JUMPING)
		Error("Jumping value too large");

	Life.currentObject->frequency = freq;
	Life.currentObject->freqLeft = freq;
	Life.currentObject->isJumpingEnabled = (freq > 1);
}


/*
 * Manipulate the tracking of old cells.
 */
static void
CmdTracks(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;

	if ((*cp == '\0') || IsAbbreviation(cp, "on"))
	{
		obj->isTracksUsed = TRUE;

		return;
	}

	if (IsAbbreviation(cp, "off"))
	{
		obj->isTracksUsed = FALSE;

		return;
	}

	if (IsAbbreviation(cp, "clear"))
	{
		ClearTracks(obj);
		ViewChanged(obj);

		return;
	}

	if (IsAbbreviation(cp, "set"))
	{
		SaveUsedCells(obj);

		ViewChanged(obj);

		return;
	}

	if (IsAbbreviation(cp, "visible"))
	{
		obj->isTracksVisible = TRUE;

		ViewChanged(obj);

		return;
	}

	if (IsAbbreviation(cp, "invisible"))
	{
		obj->isTracksVisible = FALSE;

		ViewChanged(obj);

		return;
	}

	Error("Tracks must be on, off, clear, set, visible, or invisible");
}


/*
 * Destroy an existing object.
 */
static void
CmdDestroy(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = FindObject(cp);

	if (obj == 0)
		Error("No such object");

	DestroyObject(obj);
}


/*
 * Command to create a new main window.
 */
static void
CmdWindow(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	/*
	 * Create a new top level window.
	 */
	LifeWindow * window = Life.device->openWindow(Life.device, 0);

	if (window == 0)
		Error("Failed to create window");

	/*
	 * If a parameter was given then try to load the settings
	 * and set to share those settings.
	 */
	if (cp)
	{
		UseWindowSettings(window, cp);
		LoadWindowSettings(window, cp);
	}
}


/*
 * Command to create a new sub window of the current window.
 */
static void
CmdSubWindow(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	if (Life.currentWindow->parentWindow)
		Error("Cannot create nested subwindows");

	/*
	 * Create a new subwindow.
	 */
	LifeWindow * window = Life.device->openWindow(Life.device, Life.currentWindow);

	if (window == 0)
		Error("Failed to create window");

	/*
	 * If a parameter was given then try to load the settings
	 * and set to share those settings.
	 */
	if (cp)
	{
		UseWindowSettings(window, cp);
		LoadWindowSettings(window, cp);
	}
}


/*
 * Command to close the current window.
 * The last top level window cannot be closed.
 */
static void
CmdClose(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	if (!Life.device->closeWindow(Life.device))
		Error("Cannot close last top-level window");
}


/*
 * Set the size percentage for new windows or subwindows.
 */
static void
CmdSize(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	BOOL isTopLevel = FALSE;
	int percentage;

	const char * typeName = ExtractArgument(&cp);

	if (IsAbbreviation(typeName, "toplevel"))
		isTopLevel = TRUE;
	else if (IsAbbreviation(typeName, "subwindow"))
		isTopLevel = FALSE;
	else
		Error("Size type must be toplevel or subwindow");

	percentage = GetSmallNumber(cp);

	if ((percentage < 1) || (percentage > 100))
			Error("Illegal size percentage value");

	if (isTopLevel)
		Life.topWindowSizePercentage = percentage;
	else
		Life.subWindowSizePercentage = percentage;
}


/*
 * Set a list of the viewing scale values to cycle through for an object.
 * If no arguments are given then the scale is set to the next one in the list.
 */
static void
CmdScales(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;
	LifeObject * obj = Life.currentObject;
	int scaleCount = 0;
	Scale tempScale[SCALE_LIST_SIZE];

	//
	// If there are no arguments then step to the next scale
	// in the previously set list.
	//
	if (*cp == 0)
	{
		if (obj->scaleListCount == 0)
			return;

		//
		// Advance to the next scale in the list, wrapping
		// around if necessary.
		//
		obj->scaleListIndex++;

		if (obj->scaleListIndex >= obj->scaleListCount)
			obj->scaleListIndex = 0;

		SetWindowIsAutoScale(window, FALSE);
		SetScale(window, obj->scaleList[obj->scaleListIndex]);

		AddWindowObjectUpdateFlags(obj, U_VIEW | U_STATUS);

		return;
	}

	//
	// There were some arguments.  Parse each one and add the
	// new scale value to the temporary list.
	//
	while (TRUE)
	{
		const char * arg = ExtractArgument(&cp);

		if (*arg == '\0')
			break;

		if (scaleCount >= SCALE_LIST_SIZE)
			Error("Too many scale values");

		NUMBER number = GetNumber(arg);

		if ((numCmpLong(number, MAX_SCALE) > 0) ||
			(numCmpLong(number, MIN_SCALE) < 0))
		{
			Error("Scale value out of range");
		}

		tempScale[scaleCount] = numToLong(number);
		scaleCount++;
	}

	//
	// The list is complete without error, so make it effective.
	//
	memcpy(obj->scaleList, tempScale, sizeof(tempScale));
	obj->scaleListCount = scaleCount;
	obj->scaleListIndex = 0;

	SetWindowIsAutoScale(window, FALSE);
	SetScale(window, obj->scaleList[obj->scaleListIndex]);
	AddWindowObjectUpdateFlags(obj, U_VIEW | U_STATUS);
}


/*
 * Perform an action on the settings for a window.
 */
static void
CmdSettings(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;
	int argCount;
	char ** argTable;
	const char * action;
	const char * name;

	/*
	 * Split the arguments and get the action name.
	 */
	SplitArguments(cp, &argCount, &argTable);

	if (argCount-- <= 0)
		Error("Missing settings action");

	action = *argTable++;

	/*
	 * Handle actions which don't have an argument.
	 */
	if (IsAbbreviation(action, "default"))
	{
		if (argCount > 0)
			Error("Extraneous argument");

		DefaultWindowSettings(window);

		return;
	}

	if (IsAbbreviation(action, "initialize"))
	{
		if (argCount > 0)
			Error("Extraneous argument");

		InitializeWindowSettings(window);

		return;
	}

	/*
	 * Handle settings which have an optional argument.
	 */
	if (IsAbbreviation(action, "store"))
	{
		if (argCount > 1)
			Error("Extraneous argument");

		if (argCount > 0)
			name = *argTable;
		else
			name = window->settings.sharedSettings->name;

		if (!StoreWindowSettings(window, name))
			Error("Failed to store settings");

		return;
	}

	/*
	 * Now the actions all require a single settings name.
	 */
	if (argCount-- <= 0)
		Error("Missing settings name");

	name = *argTable++;

	if (argCount > 0)
		Error("Extraneous argument");

	if (IsAbbreviation(action, "use"))
	{
		if (!UseWindowSettings(window, name))
			Error("Failed to use settings");

		return;
	}

	if (IsAbbreviation(action, "load"))
	{
		if (!LoadWindowSettings(window, name))
			Error("Settings name is not defined");

		return;
	}

	if (IsAbbreviation(action, "merge"))
	{
		if (!MergeWindowSettings(window, name))
			Error("Settings name is not defined");

		return;
	}

	/*
	 * The setting parameter is not known.
	 */
	Error("Settings must be default, initialize, load, merge, store, or use");
}


/*
 * Set some settings to be private to the window.
 */
static void
CmdPrivate(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	int argCount;
	char ** argTable;
	SettingsFlags flags;

	/*
	 * Split the arguments.
	 */
	SplitArguments(cp, &argCount, &argTable);

	/*
	 * Parse the settings list.
	 */
	flags = ParseSettingsFlags(argCount, argTable);

	if (flags == 0)
		Error("Invalid settings name");

	/*
	 * Remove the settings from the shared flags.
	 */
	SubtractSharedWindowSettings(Life.currentWindow, flags);
}


/*
 * Set some settings to be shared with other windows.
 */
static void
CmdShared(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	int argCount;
	char ** argTable;
	SettingsFlags flags;

	/*
	 * Split the arguments.
	 */
	SplitArguments(cp, &argCount, &argTable);

	/*
	 * Parse the settings list.
	 */
	flags = ParseSettingsFlags(argCount, argTable);

	if (flags == 0)
		Error("Invalid settings name");

	/*
	 * Add the settings to the shared flags.
	 */
	AddSharedWindowSettings(Life.currentWindow, flags);
}


/*
 * Show the current window settings.
 */
static void
CmdShow(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	ShowWindowSettings(Life.currentWindow);
}


/*
 * Command to set a color value.
 */
static void
CmdColor(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;
	const char * typeName = ExtractArgument(&cp);
	BOOL isSuccess = FALSE;
	int type = '\0';

	/*
	 * Get the thing to change the color of.
	 */
	if (IsAbbreviation(typeName, "background"))
		type = 'b';
	else if (IsAbbreviation(typeName, "cell"))
		type = 'c';
	else if (IsAbbreviation(typeName, "marks"))
		type = 'm';
	else if (IsAbbreviation(typeName, "tracks"))
		type = 't';
	else if (IsAbbreviation(typeName, "old"))
		type = 'o';
	else if (IsAbbreviation(typeName, "text"))
		type = 'T';
	else
		Error("Color type must be background, cell, marks, tracks, old, or text");

	while (isBlank(*cp))
		cp++;

	if (*cp == '\0')
		Error("Missing color name");

	switch (type)
	{
		case 'b':
			isSuccess = SetWindowBackgroundColor(window, cp);
			break;

		case 'c':
			isSuccess = SetWindowCellColor(window, cp);
			break;

		case 'm':
			isSuccess = SetWindowMarkColor(window, cp);
			break;

		case 'o':
			isSuccess = SetWindowOldColor(window, cp);
			break;

		case 't':
			isSuccess = SetWindowTracksColor(window, cp);
			break;

		case 'T':
			isSuccess = SetWindowTextColor(window, cp);
			break;

		default:
			isSuccess = FALSE;
			break;
	}

	if (!isSuccess)
		Error("Unknown color name");
}


/*
 * Command to set the border width.
 */
static void
CmdBorder(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	PixelCount width = GetSmallNumber(cp);

	if (width < 0)
		Error("Invalid border width");

	SetWindowBorderWidth(Life.currentWindow, width);
}


/*
 * Command to affect outline showing.
 */
static void
CmdOutlines(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;

	if (IsAbbreviation(cp, "on"))
		SetWindowIsShowOutlines(window, TRUE);
	else if (IsAbbreviation(cp, "off"))
		SetWindowIsShowOutlines(window, FALSE);
	else if (IsAbbreviation(cp, "visible"))
		SetWindowIsShowMyOutline(window, TRUE);
	else if (IsAbbreviation(cp, "invisible"))
		SetWindowIsShowMyOutline(window, FALSE);
	else
		Error("Outline action must be on, off, visible, or invisible");
}


/*
 * Command to affect the location of the status line.
 * The status window can be hidden depending on the state.
 */
static void
CmdStatus(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	StatusLocation statusLocation = 0;

	if ((*cp == '\0') || IsAbbreviation(cp, "visible"))
		statusLocation = STATUS_LOCATION_SELF;
	else if (IsAbbreviation(cp, "invisible"))
		statusLocation = STATUS_LOCATION_NONE;
	else if (IsAbbreviation(cp, "parent"))
		statusLocation = STATUS_LOCATION_PARENT;
	else
		Error("Status action must be visible, invisible, or parent");

	SetWindowStatusLocation(Life.currentWindow, statusLocation);
}


/*
 * Command to affect what location the window views.
 */
static void
CmdView(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	const char * arg;
	ViewType viewType = -1;

	arg = ExtractArgument(&cp);

	if (isLower(*arg) && (arg[1] == '\0'))
		viewType = VIEW_TYPE_WHERE_A + (*arg - 'a');
	else if (IsAbbreviation(arg, "cursor"))
		viewType = VIEW_TYPE_CURSOR;
	else if (IsAbbreviation(arg, "mouse"))
		viewType = VIEW_TYPE_MOUSE;
	else if (IsAbbreviation(arg, "self"))
		viewType = VIEW_TYPE_SELF;
	else if (IsAbbreviation(arg, "origin"))
		viewType = VIEW_TYPE_ORIGIN;
	else if (IsAbbreviation(arg, "under"))
		viewType = VIEW_TYPE_UNDER;
	else
		Error("View must be cursor, mouse, self, origin, under, or a 'where' character");

	/*
	 * Make sure nothing is left.
	 */
	arg = ExtractArgument(&cp);

	if (*arg)
		Error("Extraneous arguments");

	SetWindowViewType(Life.currentWindow, viewType);
}


/*
 * Command to affect how positioning is done when required.
 */
static void
CmdPositioning(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	PositionType positionType = -1;

	if (IsAbbreviation(cp, "center"))
		positionType = POSITION_TYPE_CENTER;
	else if (IsAbbreviation(cp, "scale"))
		positionType = POSITION_TYPE_SCALE;
	else if (IsAbbreviation(cp, "shift"))
		positionType = POSITION_TYPE_SHIFT;
	else
		Error("Positioning must be center, scale, or shift");

	SetWindowPositionType(Life.currentWindow, positionType);
}


/*
 * Command to set the debugging option value.
 */
static void
CmdDebug(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Life.debugOption = GetSmallNumber(cp);
}


/*
 * Command to specify the margin around the edges.
 */
static void
CmdMargin(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	int marginPercentage = GetSmallNumber(cp);

	if ((marginPercentage < 0) || (marginPercentage > 100))
			Error("Illegal margin value");

	SetMarginPercentage(&Life.currentWindow->settings, marginPercentage);
}


/*
 * Dump things that the user might want initialized at startup to a file.
 * These are things like the current macros and button assignments.
 */
static void
CmdDumpInitData(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	FILE *	fp;

	fp = fopen(cp, "w");

	if (fp == 0)
		Error("Cannot create file");

	WriteMacros(fp);

	Life.device->writeAssign(Life.device, fp);

	fflush(fp);

	if (ferror(fp))
	{
		fclose(fp);
		Error("Write failed");
	}

	if (fclose(fp))
		Error("Close failed");
}


/*
 * Command to advance to the next object for editing in a window.
 * The reserved objects are skipped.
 */
static void
CmdNext(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;
	LifeObject * obj = window->currentObject;

	do
	{
		obj = obj->nextObject;

		if (obj == 0)
			obj = Life.objects;
	}
	while (GetStringBufferText(&obj->name)[0] == '.');

	SetWindowObject(window, obj);

	Life.moveMode = MOVE_MODE_MOVE;
}


/*
 * Edit an object.  An empty argument implies the previous object.
 * The movement mode is also set to just moving.
 */
static void
CmdEdit(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;
	LifeObject * obj = window->prevObject;

	if (*cp)
	{
		obj = FindObject(cp);

		if (obj == 0)
			obj = GetObject(cp);
	}

	if (obj == 0)
		Error("No such object");

	SetWindowObject(window, obj);

	Life.moveMode = MOVE_MODE_MOVE;
}


/*
 * Undo one or more changes to the current object.
 * If this is the first undo, then the current object
 * is backed up so that it is not lost by the undo.
 * This can back up until we reach our undo limit.
 */
static void
CmdUndo(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Count	count;
	int	index;

	count = 1;

	if (*cp)
	{
		count = GetSmallNumber(cp);

		if (count <= 0)
			return;
	}

	if ((count >= MAX_BACKUPS) ||
		(Life.undoDepth + count >= MAX_BACKUPS))
	{
		Error("Undo depth exceeded");
	}

	/*
	 * Back up this object if the backup depth is zero.
	 */
	if (Life.undoDepth == 0)
		Backup();

	/*
	 * Calculate the index of the backup object.
	 */
	index = Life.backedUpIndex - count - Life.undoDepth;

	if (index < 0)
		index += MAX_BACKUPS;

	/*
	 * Copy the backup object into the current object.
	 */
	CopyObject(Life.backupObjects[index], Life.currentObject);

	RestoreCurrentViewParameters();

	Life.undoDepth += count;
}


/*
 * Redo one or more changes to the current object.
 * Large redo values simply restore the most recent object.
 */
static void
CmdRedo(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Count	count;
	int	index;

	count = 1;

	if (*cp)
	{
		count = GetSmallNumber(cp);

		if (count <= 0)
			return;
	}

	if (Life.undoDepth <= 0)
		Error("Already at most recent object");

	if (count > Life.undoDepth)
		count = Life.undoDepth;

	Life.undoDepth -= count;

	/*
	 * Calculate the index of the backup object to
	 * be restored.
	 */
	index = Life.backedUpIndex - Life.undoDepth;

	if (index < 0)
		index += MAX_BACKUPS;

	/*
	 * Copy the backup object into the current object.
	 */
	CopyObject(Life.backupObjects[index], Life.currentObject);

	RestoreCurrentViewParameters();
}


/*
 * End current input level.
 */
static void
CmdEndInputLevel(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	if (!IsTopInputLevel())
		Error("Cannot end top level input");

	EndInputLevel();
}


/*
 * Update the view even if inside of a loop or macro.
 */
static void
CmdUpdate(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	UpdateAllWindows();
}


/*
 * Wait until all outstanding generation computations are finished.
 */
static void
CmdWait(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	while (!Life.stop && (Life.currentObject->genLeft > 0))
	{
		DoGenerations(Life.currentObject);
		UpdateAllWindows();
	}
}


/*
 * Set output frequency for object.
 * This also disables the jumping flag.
 * Only one generation is run in each command loop, but the display
 * of a generation only occurs at the specified interval.
 */
static void
CmdFrequency(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Count freq = 1;

	if (*cp)
		freq = GetSmallExpression(cp);

	if (freq <= 0)
		Error("Illegal frequency value");

	Life.currentObject->frequency = freq;
	Life.currentObject->freqLeft = freq;
	Life.currentObject->isJumpingEnabled = FALSE;
}


/*
 * Set the mode associated with movement.
 */
static void
CmdMode(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	if ((*cp == '\0') || IsAbbreviation(cp, "move"))
	{
		Life.moveMode = MOVE_MODE_MOVE;

		return;
	}

	if (IsAbbreviation(cp, "insert"))
	{
		Life.moveMode = MOVE_MODE_INSERT;

		return;
	}

	if (IsAbbreviation(cp, "delete"))
	{
		Life.moveMode = MOVE_MODE_DELETE;

		return;
	}

	Error("Illegal movement mode");
}


/*
 * Unmark all marks which are visible.
 */
static void
CmdUnMark(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;

	Backup();

	ClearMarks(obj);
}


/*
 * Invert the marks for all cells.
 */
static void
CmdInvertMarks(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	InvertMarks(Life.currentObject);
}


/*
 * Specify a grid parameter for showing grids.
 * This is one of "off", "on", "size n", or "characters cc".
 */
static void
CmdGrid(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;
	Settings * settings = &window->settings;

	/*
	 * If there is no argument then toggle the grid flag.
	 */
	if (*cp == '\0')
	{
		BOOL isGridEnabled = GetIsGridEnabled(settings);

		SetWindowIsGridEnabled(window, !isGridEnabled);

		return;
	}

	const char * command = ExtractArgument(&cp);

	/*
	 * Now check for the various arguments.
	 */
	if (IsAbbreviation(command, "on"))
	{
		SetWindowIsGridEnabled(window, TRUE);

		return;
	}

	if (IsAbbreviation(command, "off"))
	{
		SetWindowIsGridEnabled(window, FALSE);

		return;
	}

	if (IsAbbreviation(command, "checkerboard"))
	{
		SetWindowIsGridEnabled(window, TRUE);
		SetWindowGridType(window, GRID_TYPE_CHECKERBOARD);

		return;
	}

	if (IsAbbreviation(command, "lines"))
	{
		SetWindowIsGridEnabled(window, TRUE);
		SetWindowGridType(window, GRID_TYPE_LINES);

		return;
	}

	if (IsAbbreviation(command, "dots"))
	{
		SetWindowIsGridEnabled(window, TRUE);
		SetWindowGridType(window, GRID_TYPE_DOTS);

		return;
	}

	if (IsAbbreviation(command, "characters"))
	{
		if (*cp == '\0')
		{
			SetWindowGridChar1(window, DEFAULT_GRID_CHAR_1);
			SetWindowGridChar2(window, DEFAULT_GRID_CHAR_2);

			return;
		}

		int gridChar1 = *cp++;
		int gridChar2 = gridChar1;

		if (*cp)
			gridChar2 = *cp++;

		if ((gridChar1 <= ' ') || (gridChar2 <= ' ') || *cp)
			Error("Bad grid characters");

		SetWindowGridChar1(window, gridChar1);
		SetWindowGridChar2(window, gridChar2);

		return;
	}


	if (IsAbbreviation(command, "size"))
	{
		Count size = GetSmallNumber(cp);

		if (size < 1)
			Error("Bad grid size");

		SetWindowGridSize(window, size);

		return;
	}

	Error("Invalid grid parameter");
}


/*
 * Specify what the shape used for drawing cells is.
 * This is either squares, circles, or diamonds, or large or small.
 */
static void
CmdShape(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;

	if ((*cp == '\0') || IsAbbreviation(cp, "squares"))
		SetWindowCellShape(window, SHAPE_SQUARE);
	else if (IsAbbreviation(cp, "circles"))
		SetWindowCellShape(window, SHAPE_CIRCLE);
	else if (IsAbbreviation(cp, "diamonds"))
		SetWindowCellShape(window, SHAPE_DIAMOND);
	else if (IsAbbreviation(cp, "small"))
		SetWindowIsCellShapeSmall(window, TRUE);
	else if (IsAbbreviation(cp, "large"))
		SetWindowIsCellShapeSmall(window, FALSE);
	else
		Error("Argument is 'circles', 'squares', 'diamonds', 'large' or 'small'");
}


/*
 * Specify the shifting factor for large movement.
 * The factor is 1/N of the screen width or height.
 * A value of zero restores the default shift value.
 */
static void
CmdShift(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	long shift = GetSmallExpression(cp);

	if (shift == 0)
		shift = DEFAULT_SHIFT;

	if ((shift < MIN_SHIFT) || (shift > MAX_SHIFT))
		Error("Invalid shift factor");

	SetShiftFactor(&Life.currentWindow->settings, (Scale) shift);
}


/*
 * Specify sub-windows to hide.
 * This is either our own by default, or all, or others, or none.
 */
static void
CmdHide(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeDevice * dev = Life.device;
	LifeWindow * window = Life.currentWindow;

	if ((*cp == '\0') || IsAbbreviation(cp, "self"))
	{
		if (window->parentWindow == 0)
			Error("Cannot hide top level window");

		dev->setIsVisible(dev, window, FALSE);

		return;
	}

	if (IsAbbreviation(cp, "none"))
	{
		if (window->parentWindow)
			window = window->parentWindow;

		dev->setIsVisible(dev, window, TRUE);

		return;
	}

	if (IsAbbreviation(cp, "all"))
	{
		if (window->parentWindow)
			window = window->parentWindow;

		dev->setIsVisible(dev, window, FALSE);

		return;
	}

	Error("Hide argument must be self, none, or all");
}


/*
 * Display a list of the line mode commands.
 */
static void
CmdHelp(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	const CMDTAB * cmd;
	int count;

	ShowHelp("\
The following table lists all line mode commands.  Unique abbreviations are\n\
allowed.  Commands shown with '*' can be abbreviated even when ambiguous.\n");

	count = 0;

	for (cmd = cmdtab; cmd->name; cmd++)
	{
		if (cmd->flags & F_HIDE)
			continue;

		if ((count++ % 2) == 0)
			ShowHelp("\n");

		FormatHelp("%c %-36s", ((cmd->flags & F_ABBR) ? '*' : ' '),
			cmd->name);
	}

	if (aliasCount > 0)
	{
		ShowHelp("\n\nThere ");

		if (aliasCount == 1)
			ShowHelp("is 1 alias");
		else
			FormatHelp("are %d aliases", aliasCount);

		ShowHelp(" defined (use the \"alias\" command for a list).\n");
	}

	ShowHelp("\n");
	EndHelp(TRUE);
}


/*
 * Display a list of the *.l files that can be read.
 * If a pattern is given, only list files matching that pattern.
 * If a prefix is set, then that is taken into account.
 */
static void
CmdList(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	const char *	dirname;
	int		i;
	int		len;
	List	names;
	char	buf[MAX_PATH];
	char	name[MAX_PATH];

	InitList(&names);

	/*
	 * Look for an option, which must be -d.
	 * If found, then just look in that particular directory.
	 */
	if (*cp == '-')
	{
		dirname = cp;

		while (*cp && !isBlank(*cp))
			cp++;

		if ((cp != dirname + 2) || (memcmp(dirname, "-d", 2) != 0))
			Error("Bad option for list command");

		while (isBlank(*cp))
			cp++;

		dirname = ".";

		if (*cp)
		{
			dirname = cp;

			while (*cp && !isBlank(*cp))
				cp++;

			len = cp - dirname;

			if (len >= sizeof(name))
				Error("Directory name too long");

			memcpy(name, dirname, len);
			name[len] = '\0';

			dirname = name;
		}

		while (isBlank(*cp))
			cp++;

		GetFiles(&names, dirname, cp, strlen(dirname) + 1, 1);
	}
	else if (prefix[0] == '/')
	{
		/*
		 * There was an absolute prefix set, so only look there.
		 */
		GetFiles(&names, prefix, cp, strlen(prefix) + 1, MAX_LIST);
	}
	else if (prefix[0] != '\0')
	{
		/*
		 * There was a relative prefix set.
		 * Append that path to all the library directory paths.
		 */
		GetFiles(&names, prefix, cp, strlen(prefix) + 1, 1);

		for (i = 0; i < Life.userLibCount; i++)
		{
			strcpy(buf, Life.userLibs[i]);
			strcat(buf, "/");
			strcat(buf, prefix);

			GetFiles(&names, buf, cp, strlen(buf) + 1, MAX_LIST);
		}

		strcpy(buf, LIFE_LIB);
		strcat(buf, "/");
		strcat(buf, prefix);

		GetFiles(&names, buf, cp, strlen(buf) + 1, MAX_LIST);
	}
	else
	{
		/*
		 * There is no prefix at all.
		 * Look in the library paths for files.
		 */
		GetFiles(&names, ".", cp, 2, 1);

		for (i = 0; i < Life.userLibCount; i++)
		{
			GetFiles(&names, Life.userLibs[i], cp,
				strlen(Life.userLibs[i]) + 1, MAX_LIST);
		}

		GetFiles(&names, LIFE_LIB, cp, strlen(LIFE_LIB) + 1, MAX_LIST);
	}

	if (names.argc == 0)
		Error("No files found");

	SortList(&names);

	DisplayList(&names, TRUE);

	FreeList(&names);
}


/*
 * Specify the list of locations to be shown in the view.
 * An action argument is one of 'on', 'off', 'disabled', 'enabled',
 * 'add', 'subtract', or 'clear'.
 * Some of these are followed by a single-word list of where location
 * letters (including '@' for the origin and '.' for the current location),
 * and then optional flag letters indicating what specific items are to be shown.
 */
static void
CmdLocations(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;
	const char * action;

	//
	// If there is no argument then just toggle whether or not
	// the showing of locations is enabled.
	//
	if (*cp == 0)
	{
		SetLocationsEnabled(obj, !obj->isLocationsEnabled);

		return;
	}

	//
	// First check for actions that affect all where locations.
	//
	action = ExtractArgument(&cp);

	if (IsAbbreviation(action, "on"))
	{
		SetLocationsEnabled(obj, TRUE);

		return;
	}

	if (IsAbbreviation(action, "off"))
	{
		SetLocationsEnabled(obj, FALSE);

		return;
	}

	if (IsAbbreviation(action, "clear"))
	{
		ClearAllLocationInfo(obj);
		SetLocationsEnabled(obj, FALSE);

		return;
	}

	//
	// Now check for actions which apply to listed where locations.
	//
	if (IsAbbreviation(action, "add"))
	{
		SetWhereLocationsShown(obj, cp, TRUE);

		return;
	}

	if (IsAbbreviation(action, "subtract"))
	{
		SetWhereLocationsShown(obj, cp, FALSE);

		return;
	}

	if (IsAbbreviation(action, "enabled"))
	{
		SetWhereLocationsEnabled(obj, cp, TRUE);

		return;
	}

	if (IsAbbreviation(action, "disabled"))
	{
		SetWhereLocationsEnabled(obj, cp, FALSE);

		return;
	}

	Error("Action is on, off, clear, disabled, enabled, add, or subtract");
}


/*
 * Set whether the specified where locations are enabled to be shown.
 * The letters are a string of 'a' to 'z', '@', or '.'.
 */
static void
SetWhereLocationsEnabled(LifeObject * obj, const char * cp, BOOL isEnabled)
{
	const char * ids = ExtractArgument(&cp);

	CheckLocationList(ids);

	//
	// Now set the enabled flag for the locations.
	//
	while (*ids)
	{
		LocationInfo * info = GetLocationInfo(obj, *ids++);

		info->isEnabled = isEnabled;
	}

	if (isEnabled)
		SetLocationsEnabled(obj, TRUE);

	AddWindowObjectUpdateFlags(obj, U_REDRAW);
}


/*
 * Set whether the where locations to be shown.
 * The letters are a string of 'a' to 'z', '@', or '.'.
 * These can optionally be followed by a string of values
 * from the set fdhvdc.
 */
static void
SetWhereLocationsShown(LifeObject * obj, const char * cp, BOOL isShown)
{
	const char * ids = ExtractArgument(&cp);
	const char * test;

	CheckLocationList(ids);

	//
	// Now get a possible set of items to affect.
	// Default to just specifying to show the id character.
	//
	if (*cp == '\0')
		cp = "i";

	for (test = cp; *test; test++)
	{
		if (strchr("ifbrcd", *test) == NULL)
			Error("Items must be in \"ifbrcd\"");
	}

	/*
	 * Now set the flags for the list of locations.
	 */
	while (*ids)
	{
		LocationInfo * info = GetLocationInfo(obj, *ids++);

		for (test = cp; *test; test++)
		{
			switch (*test)
			{
				case 'i': info->isIdShown = isShown; break;
				case 'f': info->isForwShown = isShown; break;
				case 'b': info->isBackShown = isShown; break;
				case 'r': info->isHorizShown = isShown; break;
				case 'c': info->isVertShown = isShown; break;
				case 'd': info->isDistShown = isShown; break;
				default: break;
			}
		}

		if (isShown)
			info->isEnabled = TRUE;
	}

	if (isShown)
		SetLocationsEnabled(obj, TRUE);

	AddWindowObjectUpdateFlags(obj, U_REDRAW);
}


/*
 * Determine whether the specified list of characters are valid location ids.
 * An error is generated if the list is empty or invalid.
 * These are the lower case letters, '@', or '.'.
 */
static void
CheckLocationList(const char * str)
{
	if (*str == '\0')
		Error("String of where locations expected");

	while (*str)
	{
		int id = *str++;

		if (!IS_LOCINFO_ID(id))
			Error("Only letters, '@', or '.' allowed in string");
	}
}


/*
 * Return the location info for the specified id.
 * The id is a lower case letter, '@', or '.'.
 * Fails if the id is invalid.
 */
static LocationInfo *
GetLocationInfo(LifeObject * obj, int id)
{
	if (id == LOCINFO_ID_ORIGIN)
		return &obj->originInfo;

	if (id == LOCINFO_ID_CURRENT)
		return &obj->currentInfo;

	if ((id >= 'a') && (id <= 'z'))
		return &obj->whereInfo[id - 'a'];

	Error("Invalid location id");

	return 0;
}


/*
 * Set whether or not the where locations showing is enabled.
 */
static void
SetLocationsEnabled(LifeObject * obj, BOOL isEnabled)
{
	if (obj->isLocationsEnabled != isEnabled)
	{
		obj->isLocationsEnabled = isEnabled;

		AddWindowObjectUpdateFlags(obj, U_REDRAW);
	}
}


/*
 * Insert another object into this one.
 * This preserves existing cells in the current object.
 */
static void
CmdInsert(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = FindObject(cp);

	if (obj == 0)
		Error("No such object");

	Backup();

	AddObject(obj, Life.currentObject, ADD_RELATIVE);
}


/*
 * Show a list of objects.
 */
static void
CmdObjects(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	BOOL isAll = ((*cp == '-') || (*cp == 'a'));

	ListObjects(isAll);
}


/*
 * Show information about the specified object, or the current one.
 */
static void
CmdInformation(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	const LifeObject * obj = Life.currentObject;

	if (*cp)
	{
		obj = FindObject(cp);

		if (obj == 0)
			Error("No such object");
	}

	ShowObjectInfo(obj);
}


/*
 * Show the quad cell statistics.
 */
static void
CmdQuadCell(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	/*
	 * Show the statistics by default.
	 */
	if ((*cp == '\0') || (strcmp(cp, "show") == 0))
	{
		ShowQuadCellStatistics();
		return;
	}

	if (strcmp(cp, "clear") == 0)
	{
		qcClearStats();

		return;
	}

	Error("Invalid quadcell action");
}


/*
 * Rename the current object to something else.
 */
static void
CmdRename(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	if (Life.currentObject->isReserved)
		Error("Cannot rename reserved object");

	if (*cp == '\0')
		Error("Empty object name");

	if (strchr(cp, ' ') || strchr(cp, '\t'))
		Error("Object name contains blanks");

	if (FindObject(cp))
		Error("Name already exists");

	if (BAD_NAME(cp))
		Error("Cannot create reserved name");

	SetStringBuffer(&Life.currentObject->name, cp);
}

/*
 * Overlay the view of another object on this one.
 * If no object is given then any previous one is toggled active.
 */
static void
CmdOverlay(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;

	//
	// If there is no object name then toggle the old one.
	//
	if (*cp == '\0')
	{
		if (obj->overlayObject == 0)
			Error("No previous overlay object");

		obj->isOverlayActive = !obj->isOverlayActive;

		return;
	}

	//
	// A new overlay object is being set.
	//
	if (strchr(cp, ' ') || strchr(cp, '\t'))
		Error("Object name contains blanks");

	LifeObject * overlayObject = FindObject(cp);

	if (overlayObject == 0)
		Error("Overlay object does not exist");

	if (overlayObject == obj)
		Error("Cannot overlay same object");

	obj->overlayObject = overlayObject;
	obj->isOverlayActive = TRUE;
}


/*
 * Move current object into another object.
 */
static void
CmdMove(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	MoveObject(Life.currentObject, GetObject(cp));
}


/*
 * Move currently marked cells into another object.
 */
static void
CmdMoveMarked(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	if (!SelectMarkedCells(Life.currentObject))
		return;

	MoveSelection(Life.currentObject, GetObject(cp));
}


/*
 * Set the value of a variable.
 */
static void
CmdSet(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	const char *	var;

	var = ExtractArgument(&cp);

	if (*var == '\0')
		Error("Missing variable");

	/*
	 * If there is no expression, then set the variable to zero.
	 */
	if (*cp == '\0')
	{
		SetVariable(var, numZero());

		return;
	}

	if (*cp == '=')
		cp++;

	SetVariable(var, GetExpression(cp));
}


/*
 * Type the value of an expression.
 */
static void
CmdType(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	static char buf[128];

	sprintf(buf, "Value = %s", numToString(GetExpression(cp)));
	Message(buf);
}


/*
 * Set the delay time between updates in milliseconds.
 */
static void
CmdDelay(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	long	milliSeconds;

	if (*cp == '\0')
	{
		Life.delayTime = 0;

		return;
	}

	milliSeconds = GetSmallExpression(cp);

	if ((milliSeconds < 0) || (milliSeconds > MAX_DELAY))
		Error("Illegal delay time");

	Life.delayTime = milliSeconds;
}


/*
 * Display the values of all the variables.
 */
static void
CmdVariables(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	ListVariables();
}


/*
 * Quit program
 */
static void
CmdQuit(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	Life.device->closeDevice(Life.device);

	exit(0);
}


/*
 * Set or display the path prefix for reading and listing files.
 */
static void
CmdPrefix(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	int		len;
	static	char	buf[MAX_PREFIX + 20];

	/*
	 * If there is no argument then display the current prefix.
	 */
	if (*cp == '\0')
	{
		if (prefix[0])
		{
			sprintf(buf, "Prefix = %s", prefix);
			Message(buf);
		}
		else
			Message("No prefix");

		return;
	}

	/*
	 * If the prefix is just a dot then clear it.
	 */
	if ((cp[0] == '.') && (cp[1] == '\0'))
	{
		prefix[0] = '\0';

		return;
	}

	/*
	 * If the prefix is too long then complain.
	 */
	len = strlen(cp);

	if (len >= MAX_PREFIX)
		Error("Prefix string is too long");

	/*
	 * Remove trailing slashes except for an initial one,
	 * and then save the prefix away for future use.
	 */
	while ((len > 1) && (cp[len - 1] == '/'))
		len--;

	memcpy(prefix, cp, len);
	prefix[len] = '\0';
}


/*
 * Randomize a square area centered on the cursor with cells which
 * are on the specified percentage of the time.
 */
static void
CmdRandomize(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeObject * obj = Life.currentObject;
	long	chance;
	Coord	row;
	Coord	col;

	chance = RANDOMCHANCE;

	if (*cp)
		chance = GetSmallExpression(cp);

	if ((chance < 0) || (chance > 100))
		Error("Randomizing chance must be between 0 and 100");

	Backup();

	for (row = numToLong(obj->currentInfo.row) - RANDOMRADIUS;
		row <= numToLong(obj->currentInfo.row) + RANDOMRADIUS; row++)
	{
		for (col = numToLong(obj->currentInfo.col) - RANDOMRADIUS;
			col <= numToLong(obj->currentInfo.col) + RANDOMRADIUS; col++)
		{
			if (Life.stop)
				return;

			if ((random() % 100) < chance)
			{
				AddCell(obj, numFromLong(row),
					numFromLong(col));
			}
			else
			{
				DeleteCell(obj, numFromLong(row),
					numFromLong(col));
			}
		}
	}
}


/*
 * Load an object from a file, defaulting extension if needed.
 * If a prefix is set, then apply that prefix to the name.
 * The object is loaded into a new buffer having the file's name,
 * or one close to that.  Multiple objects can be given in one command.
 */
static void
CmdLoad(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	while (TRUE)
	{
		char	buf[MAX_PATH];

		const char * path = ExtractArgument(&cp);

		if (*path == 0)
			return;

		/*
		 * We are to read a normal file.
		 * If there is a prefix then apply it.
		 */
		if (prefix[0])
		{
			if (strlen(prefix) + strlen(path) + 2 > MAX_PATH)
				Error("Path is too long");

			strcpy(buf, prefix);

			if ((prefix[0] != '/') || (prefix[1] != '\0'))
				strcat(buf, "/");

			strcat(buf, path);

			path = buf;
		}

		//
		// Look for the file in our libraries with our extensions.
		//
		path = FindLibraryFile(path);

		if (path == 0)
			Error("File not found");

		LoadFile(path);
	}
}


/*
 * Read commands or object from file, defaulting extension if needed.
 * If a prefix is set, then apply that prefix to the name.
 * The special parameter of "-inline" means that an RLE or XLIFE
 * format object is to be read from the current input file.
 */
static void
CmdRead(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	FILE *	fp;
	char	buf[MAX_PATH];

	Backup();

	/*
	 * If this is the inline command then read a formatted object
	 * coming right after this command instead of reading the
	 * object from a file name.  This can only be used if the
	 * current input source is from a file.
	 */
	if (strcmp(cp, "-inline") == 0)
	{
		fp = GetInputFILE();

		if (fp == 0)
			Error("Cannot use inline read outside of a file");

		if (!TryReadFile(fp, Life.currentObject))
			Error("Invalid data for inline read");

		return;
	}

	/*
	 * We are to read a normal file.
	 * If there is a prefix then apply it.
	 */
	if (prefix[0])
	{
		if (strlen(prefix) + strlen(cp) + 2 > MAX_PATH)
			Error("Path is too long");

		strcpy(buf, prefix);

		if ((prefix[0] != '/') || (prefix[1] != '\0'))
			strcat(buf, "/");

		strcat(buf, cp);

		cp = buf;
	}

	if (!SetFile(cp))
		Error("Cannot open input file");
}


/*
 * Set new life rule.
 */
static void
CmdRule(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	if (*cp == '\0')
		cp = DEFAULT_RULE;

	Life.currentObject->rule = ParseRule(cp);
}


/*
 * Read commands from the terminal.  Useful in loops or command files.
 * If the -c argument is given, we don't do it if no input is ready.
 */
static void
CmdTtyInput(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	/*
	 * Possibly check for input.
	 */
	if ((*cp == '-') || (*cp == 'c'))
	{
		if (!Life.device->isInputReady(Life.device))
			return;
	}

	if (!SetTerminal())
		Error("Nesting too deep");
}


/*
 * Write object to file.
 * The format can be specified as an option before the file name.
 */
static void
CmdWrite(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	const char * method = 0;
	int len;
	char name[16];

	if (!got1)
		arg1 = numFromLong(WRITECOLS);

	if (*cp == '-')
	{
		method = cp + 1;

		while (*cp && !isBlank(*cp))
			cp++;

		len = cp - method;

		if (len >= sizeof(name))
			Error("Method name too long");

		memcpy(name, method, len);
		name[len] = '\0';

		method = name;

		while (isBlank(*cp))
			cp++;
	}

	/*
	 * See if input is from the terminal and whether the output file
	 * exists.  If so, ask confirmation about overwriting the file.
	 */
	if (IsTerminalInput() && (access(cp, F_OK) == 0))
	{
		if (!Confirm("Overwrite existing file? ", FALSE))
			Error("Write aborted");
	}

	WriteObject(Life.currentObject, method, cp,
		(numIsNonZero(arg1) ? WRITE_ROWS : 0), numToLong(arg1));
}


/*
 * Zero out current object.
 */
static void
CmdZero(const char * cp, NUMBER arg1, NUMBER arg2, BOOL got1, BOOL got2)
{
	LifeWindow * window = Life.currentWindow;
	LifeObject * obj = Life.currentObject;

	Backup();

	ZeroObject(obj);

	InitAllLocationInfo(obj);

	obj->stopGap = 0;
	obj->stopOnPath = FALSE;
	obj->stopInsidePath = FALSE;
	obj->stopOutsidePath = FALSE;
	obj->stopCell = FALSE;
	obj->stopState = FALSE;
	obj->stopRow = numZero();
	obj->stopCol = numZero();
	obj->stopExpr[0] = '\0';

	SetScale(Life.currentWindow, GetDefaultScale(&window->settings));

	obj->frequency = Life.defaultFrequency;
	obj->isJumpingEnabled = Life.defaultIsJumpingEnabled;
	obj->freqLeft = Life.defaultFrequency;
	obj->isTracksUsed = Life.defaultTracks;
	obj->isTracksVisible= TRUE;

	Life.moveMode = MOVE_MODE_MOVE;
}


/*
 * See if one string is an abbreviation of another.
 * Returns TRUE if first string is an abbreviation.
 * However, an empty string is never an abbreviation.
 */
BOOL
IsAbbreviation(const char * str1, const char * str2)
{
	if ((str1 == 0) || (str2 == 0))
		return FALSE;

	if (*str1 == '\0')
		return FALSE;

	while (*str1)
	{
		if (*str1++ != *str2++)
			return FALSE;
	}

	return TRUE;
}


/*
 * Parse a small number from a null-terminated string.
 * The number can be surrounded by blanks but otherwise
 * it must be valid with no extra characters.
 */
static long
GetSmallNumber(const char * cp)
{
	NUMBER value = GetNumber(cp);

	if (!numIsLong(value))
		Error("Number is larger than a long");

	return numToLong(value);
}


/*
 * Parse a number from a null-terminated string.
 * The number can be surrounded by blanks but otherwise
 * it must be valid with no extra characters.
 */
static NUMBER
GetNumber(const char * cp)
{
	NUMBER result = numZero();
	BOOL isNegative = FALSE;

	while (isBlank(*cp))
		cp++;

	if (*cp == '\0')
		Error("Missing number");

	if (*cp == '-')
	{
		isNegative = TRUE;
		cp++;
	}

	if (!isDigit(*cp))
		Error("Invalid number");

	while (isDigit(*cp))
		result = numAddLong(numMulLong(result, 10), *cp++ - '0');

	while (isBlank(*cp))
		cp++;

	if (*cp || numIsNeg(result))
		Error("Invalid number");

	if (isNegative)
		result = numNeg(result);

	return result;
}


/*
 * Set the view changed flag for the specified object.
 * This also sets that the status has changed since that is
 * generally the case.
 */
static void
ViewChanged(const LifeObject * obj)
{
	AddWindowObjectUpdateFlags(obj, U_VIEW | U_STATUS);
}


/* END CODE */
