/*
 * Copyright (c) 2018 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Include file for view info data and routines.
 */
#ifndef	VIEWINFO_H
#define	VIEWINFO_H

#include "number.h"


/*
 * View ids for different uses in the life program.
 */
#define	VIEW_ID_PRIVATE	-1
#define	VIEW_ID_OBJECT	0
#define	VIEW_ID_MARKS	1
#define	VIEW_ID_TRACKS	2
#define	VIEW_ID_OVERLAY	3


/*
 * Information about a rectangular region of cells which is being examined.
 * Routines can scan the region row by row to find interesting cells.
 */
typedef struct
{
	char * data;		// data array containing just zeros and ones
	char * auxData;		// auxillary data array or 0
	int dataSize;		// size of either data array
	int rowCount;		// number of rows
	int colCount;		// number of columns
	int id;			// identifier for data allocation (or -1)
	int scale;		// scale factor of the view
	BOOL isValid;		// whether the view has been inited
	NUMBER onCount;		// number of set cells in data
	NUMBER minRow;		// minimum row coordinate of data
	NUMBER maxRow;		// maximum row coordinate of data
	NUMBER minCol;		// minimum column coordinate of data
	NUMBER maxCol;		// maximum column coordinate of data

	//
	// The values which change during use.
	//
	int rowIndex;		// row index of the current row
	int dataOffset;		// current offset into the data
	int endRowOffset;	// offset of the end of the current row
} ViewInfo;


//
// Routines which allow a view of a limited region of the grid.
//
extern void InitView(ViewInfo * info, int id, BOOL isAuxUsed, int scale,
	NUMBER minRow, NUMBER maxRow, NUMBER minCol, NUMBER maxCol);

extern void InitEmptyView(ViewInfo * info);

extern void FreeView(ViewInfo * info);
extern void ResetView(ViewInfo * info);

extern int FindNextRowIndex(ViewInfo * info);
extern int FindNextColIndex(ViewInfo * info);
extern int PickMinimumIndex(int index1, int index2, int index3, int index4);

extern BOOL AreViewDataBitsSet(const ViewInfo * info,
	int minRow, int maxRow, int minCol, int maxCol, int bitValue);

extern BOOL IsViewCellSet(const ViewInfo * info, int rowIndex, int colIndex);
extern void SetViewCell(ViewInfo * info, int rowIndex, int colIndex, BOOL isSet);
extern int GetAuxViewData(const ViewInfo * info, int rowIndex, int colIndex);
extern void SetAuxViewData(ViewInfo * info, int rowIndex, int colIndex, int value);
extern void OrAuxViewData(ViewInfo * info, int rowIndex, int colIndex, int value);
extern void FillViewData(ViewInfo * info, int value);
extern void FillAuxViewData(ViewInfo * info, int value);
extern void ClearViewData(ViewInfo * info);

#endif

